// Copyright (C) 2003 by Alexander Motzkau.
// 
// This file is released under the GPL.

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <libdevmapper.h>

#include "sha256.h"

// Because we should not include kernel headers like <linux/fs.h>
#define BLKGETSIZE _IO(0x12,96)

char *hex = "0123456789abcdef";

struct device_list
{
    unsigned char* cipher;
    unsigned char* device;
    unsigned long start;
    unsigned long size;
    struct device_list* next;
};

// We build an AVL tree.
struct task_list
{
    struct task_list* childs[2];
    int balance; // -1: left ([0]) is heigher, 0: balanced,
		 //  1: right ([1]) is heigher
    unsigned char* name;
    struct device_list* devices;
};

char* make_string(char* str)
{
    int n;
    char* nstr;
    
    n = strlen(str)+1;
    nstr = (char*) malloc(n);
    return strncpy(nstr, str, n);
}

#define min(a,b) (((a)>(b))?(b):(a))
#define max(a,b) (((a)>(b))?(a):(b))

void rotate(struct task_list** root, int idx)
{
    /*
          oldroot                     newroot
	 /       \                   /       \
	/	  \                 /         \
		newroot     ->  oldroot
	       /      \        /       \
              /        \      /         \
	      
height: a    b+1      c+1   a+1        b+1     c

Old Balances: oldroot: max(b,c)-a+1
	      newroot: c-b
New Balances: oldroot': b-a = oldroot-1-max(newroot,0)
	      newroot': c-max(a,b)-1 = newroot-1+min(oldroot',0)

         oldroot               newroot
	/       \             /       \
    newroot           ->            oldroot
   /       \                       /       \

  c+1     b+1    a          c     b+1     a+1

Old Bal: oldroot: -max(b,c)+a-1
         newroot: b-c
New Bal: oldroot': a-b =  oldroot+1-min(newroot,0)
	 newroot: max(a,b)+1-c =  newroot+1+max(oldroot',0)
    */

    struct task_list* oldroot, *newroot;
    
    oldroot = (*root);
    newroot = oldroot->childs[idx];
    
    oldroot->childs[idx] = newroot->childs[1-idx];
    newroot->childs[1-idx] = oldroot;
    *root = newroot;

    if(idx)
    {
	oldroot->balance -= max(newroot->balance,0)+1;
	newroot->balance += min(oldroot->balance,0)-1;
    }
    else
    {
	oldroot->balance -= min(newroot->balance,0)-1;
	newroot->balance += max(oldroot->balance,0)+1;
    }
}

// Returns 1, if the height of the subtree increased, 0 otherwise.
int insert(struct task_list* tasks, char* name, struct device_list* device, struct task_list** newroot)
{
    int cmp;

    if(tasks==NULL)
    {
	*newroot = (struct task_list*) malloc(sizeof(struct task_list));
	(*newroot)->childs[0] = NULL;
	(*newroot)->childs[1] = NULL;
	(*newroot)->balance = 0;
	(*newroot)->name = make_string(name);
	(*newroot)->devices = device;
	return 1;
    }
    
    cmp = strcmp(name, tasks->name);
    if(cmp==0)
    {
	struct device_list** current;
	
	current = &(tasks->devices);
	while(*current!=NULL)
	    current = &((*current)->next);
	*current = device;
	return 0;
    }
    else
    {
	int idx;
	if(cmp<0)
	{
	    cmp = -1;
	    idx = 0;
	}
	else
	{
	    cmp = 1;
	    idx = 1;
	}

	if(insert(tasks->childs[idx], name, device, &tasks->childs[idx]))
	{
	    if(tasks->balance!=cmp)
	    {
		tasks->balance += cmp;
		return (tasks->balance!=0);
	    }
	    else
	    {
		tasks->balance += cmp;
		// We have to rotate this one...
		
		if(tasks->childs[idx]->balance==-cmp) // Double Rotate?
		    rotate(&(tasks->childs[idx]), 1-idx);
		
		rotate(newroot, idx);
		return 0;
	    }
	}
	
	return 0;
    }
}

struct task_list* read_table(const char* tablefile)
{
    int line;
    FILE *table;
    struct task_list* root;

    root = NULL;
    
    table = fopen(tablefile,"rt");
    if(table==NULL)
    {
	perror(tablefile);
	return NULL;
    }
    
    // Now read the table
    line = 0;
    while(!feof(table))
    {
	unsigned char buf[1024];
	char* ptr;
	char* name;
	char* cipher;
	char* device;
	unsigned long start, size;
	int i;
	struct device_list* devstr;
		
	line++;
	if(fgets(buf, 1024, table)==NULL)
	    break;
	ptr=buf;
	
	while(isspace(*ptr))
	    ptr++;
	if(!*ptr || *ptr=='#')
	    continue;
	
#define NEXT_ELEMENT 		\
	while(isgraph(*ptr))	\
	    ptr++;		\
	if(*ptr)		\
	{			\
	    *ptr=0;		\
	    ptr++;		\
	    while(isspace(*ptr))\
		ptr++;		\
	}
	
	name=ptr;   NEXT_ELEMENT;
	cipher=ptr; NEXT_ELEMENT;
	device=ptr; NEXT_ELEMENT;

	if(ptr==device)
	{
	    fprintf(stderr, "%s line %d: Invalid format.\n", tablefile, line);
	    fclose(table);
	    return NULL;
	}
	if(*ptr)
	    i=sscanf(ptr,"%lu %lu", &start, &size);
	else
	    i=0;

	if(!i || i<0)
	    start=0;

	if(i<2)
	{
	    FILE *dev;
	    dev = fopen(device, "rb");
	    if(dev==NULL || ioctl(fileno(dev),BLKGETSIZE, &size)==-1)
	    {
		perror(device);
		fclose(dev);
		fclose(table);
		return NULL;
	    }
	    fclose(dev);
	}
	
	devstr = (struct device_list*) malloc(sizeof(struct device_list));
	devstr->cipher = make_string(cipher);
	devstr->device = make_string(device);
	devstr->start = start;
	devstr->size = size;
	devstr->next = NULL;
	
	insert(root, name, devstr, &root);
    }        

    return root;
}

int get_key(char* keyfile, unsigned char* key)
{
    unsigned char buf[1024];
    unsigned char pwkey[32];
    sha256_context sctx;
    int i;
    
    // Get the key file, if present.
    if(keyfile!=NULL)
    {
	FILE *inkey;
	inkey = fopen(keyfile, "rb");
	if(inkey==NULL)
	{
	    perror(keyfile);
	    return 1;
	}
	
	if(!fread(key, 32, 1, inkey))
	{
	    fprintf(stderr, "%s: Too small. 32 Bytes needed.\n", keyfile);
	    fclose(inkey);
	    return 1;
	}
	fclose(inkey);
    }
    else
	memset(key, 0, 32);
    
    // Let's get the password.
    //fprintf(stdout,"Passphrase (end with Ctrl+D): "); fflush(stdout);
    sha256_starts( &sctx);
    while( (i=fread(buf, 1, sizeof(buf), stdin))>0)
	sha256_update(&sctx, buf, i);
    sha256_finish(&sctx, pwkey);
    
    for(i=0;i<32;i++)
	key[i]^=pwkey[i];
    
    return 0;
}

void crypt_fs(struct task_list* tasks, unsigned char* key)
{
    if(tasks==NULL)
	return;
	
    crypt_fs(tasks->childs[0], key);
    
    do
    {
	struct device_list* dev;
	struct dm_task* dmt;
	unsigned long pos;

	// initialize device mapper.    
        if(!(dmt=dm_task_create(DM_DEVICE_CREATE)))
	{
	    fprintf(stderr, "%s: Couldn't create device mapper task.\n", tasks->name);
	    break;
	}
    
	if(!dm_task_set_name(dmt, tasks->name))
	{
	    dm_task_destroy(dmt);
	    fprintf(stderr, "%s: Couldn't set task name.\n", tasks->name);
	    break;
	}
    
	pos = 0;
	dev = tasks->devices;

	while(dev!=NULL)
	{
	    char pars[256];
	    int i, j;
	    
	    strncpy(pars, dev->cipher, 256);
	    i = strlen(pars);
	    
	    if(i+35>256)
	    {
		fprintf(stderr, "%s: cipher name too long.\n", tasks->name);
		dm_task_destroy(dmt);
		pos=-1;
		break;
	    }
	    
	    pars[i++]=' ';
	    for(j=0;j<32;j++)
	    {
		int k = key[j];
	        pars[i++]=hex[k>>4];
		pars[i++]=hex[k&15];
	    }
	    pars[i++]=' ';
	    
	    if(snprintf(pars+i, 256-i, "%lu %s %lu", pos, dev->device, dev->start)>255-i)
	    {
		fprintf(stderr, "%s: line too long.\n", tasks->name);
		dm_task_destroy(dmt);
		pos=-1;
		break;
	    }

	    if (!dm_task_add_target(dmt, pos, dev->size, "crypt", pars))
	    {
		dm_task_destroy(dmt);
		fprintf(stderr, "%s: couldn't add cipher.\n", tasks->name);
		pos=-1;
		break;
	    }
	    pos+= dev->size;
	    dev = dev->next;
	}
	if(pos<0)
	    break;
    
	if (!dm_task_run(dmt))
        {
	    dm_task_destroy(dmt);
	    fprintf(stderr, "%s: Couldn't run task.\n", tasks->name);
	    break;
	}
    }
    while(0);
    
    crypt_fs(tasks->childs[1], key);
}

int usage(FILE* out)
{
    fprintf(out, "Usage: cryptfs <device-table> [<key-file>]\n");
    return 1;
}

int main(int argc, char *argv[])
{
    unsigned char key[32];
    struct task_list* tasks;
    
    if(argc<2 || argc>3)
	return usage(stderr);

    tasks = read_table(argv[1]);
    if(tasks==NULL)
	return 1;

    if(get_key((argc==3)?argv[2]:NULL, key))
	return 1;

    crypt_fs(tasks, key);

    dm_lib_release();
    dm_lib_exit();
    return 0;
}
